@@ -1,6 +1,6 @@
 /* GNU m4 -- A simple macro processor
 
-   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2000, 2004, 2006
+   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2000, 2004, 2006, 2007
    Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -19,7 +19,7 @@
    02110-1301  USA
 */
 
-/* Code for all builtin macros, initialisation of symbol table, and
+/* Code for all builtin macros, initialization of symbol table, and
    expansion of user defined macros.  */
 
 #include "m4.h"
@@ -34,8 +34,8 @@ extern FILE *popen ();
 
 #define ARG(i)	(argc > (i) ? TOKEN_DATA_TEXT (argv[i]) : "")
 
-/* Initialisation of builtin and predefined macros.  The table
-   "builtin_tab" is both used for initialisation, and by the "builtin"
+/* Initialization of builtin and predefined macros.  The table
+   "builtin_tab" is both used for initialization, and by the "builtin"
    builtin.  */
 
 #define DECLARE(name) \
@@ -73,6 +73,7 @@ DECLARE (m4_len);
 DECLARE (m4_m4exit);
 DECLARE (m4_m4wrap);
 DECLARE (m4_maketemp);
+DECLARE (m4_mkstemp);
 DECLARE (m4_patsubst);
 DECLARE (m4_popdef);
 DECLARE (m4_pushdef);
@@ -96,59 +97,60 @@ builtin_tab[] =
 
   /* name		GNUext	macros	blind	function */
 
-  { "__file__",		TRUE,	FALSE,	FALSE,	m4___file__ },
-  { "__line__",		TRUE,	FALSE,	FALSE,	m4___line__ },
-  { "__program__",	TRUE,	FALSE,	FALSE,	m4___program__ },
-  { "builtin",		TRUE,	FALSE,	TRUE,	m4_builtin },
-  { "changecom",	FALSE,	FALSE,	FALSE,	m4_changecom },
-  { "changequote",	FALSE,	FALSE,	FALSE,	m4_changequote },
+  { "__file__",		true,	false,	false,	m4___file__ },
+  { "__line__",		true,	false,	false,	m4___line__ },
+  { "__program__",	true,	false,	false,	m4___program__ },
+  { "builtin",		true,	true,	true,	m4_builtin },
+  { "changecom",	false,	false,	false,	m4_changecom },
+  { "changequote",	false,	false,	false,	m4_changequote },
 #ifdef ENABLE_CHANGEWORD
-  { "changeword",	TRUE,	FALSE,	TRUE,	m4_changeword },
+  { "changeword",	true,	false,	true,	m4_changeword },
 #endif
-  { "debugmode",	TRUE,	FALSE,	FALSE,	m4_debugmode },
-  { "debugfile",	TRUE,	FALSE,	FALSE,	m4_debugfile },
-  { "decr",		FALSE,	FALSE,	TRUE,	m4_decr },
-  { "define",		FALSE,	TRUE,	TRUE,	m4_define },
-  { "defn",		FALSE,	FALSE,	TRUE,	m4_defn },
-  { "divert",		FALSE,	FALSE,	FALSE,	m4_divert },
-  { "divnum",		FALSE,	FALSE,	FALSE,	m4_divnum },
-  { "dnl",		FALSE,	FALSE,	FALSE,	m4_dnl },
-  { "dumpdef",		FALSE,	FALSE,	FALSE,	m4_dumpdef },
-  { "errprint",		FALSE,	FALSE,	TRUE,	m4_errprint },
-  { "esyscmd",		TRUE,	FALSE,	TRUE,	m4_esyscmd },
-  { "eval",		FALSE,	FALSE,	TRUE,	m4_eval },
-  { "format",		TRUE,	FALSE,	TRUE,	m4_format },
-  { "ifdef",		FALSE,	FALSE,	TRUE,	m4_ifdef },
-  { "ifelse",		FALSE,	FALSE,	TRUE,	m4_ifelse },
-  { "include",		FALSE,	FALSE,	TRUE,	m4_include },
-  { "incr",		FALSE,	FALSE,	TRUE,	m4_incr },
-  { "index",		FALSE,	FALSE,	TRUE,	m4_index },
-  { "indir",		TRUE,	FALSE,	TRUE,	m4_indir },
-  { "len",		FALSE,	FALSE,	TRUE,	m4_len },
-  { "m4exit",		FALSE,	FALSE,	FALSE,	m4_m4exit },
-  { "m4wrap",		FALSE,	FALSE,	TRUE,	m4_m4wrap },
-  { "maketemp",		FALSE,	FALSE,	TRUE,	m4_maketemp },
-  { "patsubst",		TRUE,	FALSE,	TRUE,	m4_patsubst },
-  { "popdef",		FALSE,	FALSE,	TRUE,	m4_popdef },
-  { "pushdef",		FALSE,	TRUE,	TRUE,	m4_pushdef },
-  { "regexp",		TRUE,	FALSE,	TRUE,	m4_regexp },
-  { "shift",		FALSE,	FALSE,	TRUE,	m4_shift },
-  { "sinclude",		FALSE,	FALSE,	TRUE,	m4_sinclude },
-  { "substr",		FALSE,	FALSE,	TRUE,	m4_substr },
-  { "syscmd",		FALSE,	FALSE,	TRUE,	m4_syscmd },
-  { "sysval",		FALSE,	FALSE,	FALSE,	m4_sysval },
-  { "traceoff",		FALSE,	FALSE,	FALSE,	m4_traceoff },
-  { "traceon",		FALSE,	FALSE,	FALSE,	m4_traceon },
-  { "translit",		FALSE,	FALSE,	TRUE,	m4_translit },
-  { "undefine",		FALSE,	FALSE,	TRUE,	m4_undefine },
-  { "undivert",		FALSE,	FALSE,	FALSE,	m4_undivert },
+  { "debugmode",	true,	false,	false,	m4_debugmode },
+  { "debugfile",	true,	false,	false,	m4_debugfile },
+  { "decr",		false,	false,	true,	m4_decr },
+  { "define",		false,	true,	true,	m4_define },
+  { "defn",		false,	false,	true,	m4_defn },
+  { "divert",		false,	false,	false,	m4_divert },
+  { "divnum",		false,	false,	false,	m4_divnum },
+  { "dnl",		false,	false,	false,	m4_dnl },
+  { "dumpdef",		false,	false,	false,	m4_dumpdef },
+  { "errprint",		false,	false,	true,	m4_errprint },
+  { "esyscmd",		true,	false,	true,	m4_esyscmd },
+  { "eval",		false,	false,	true,	m4_eval },
+  { "format",		true,	false,	true,	m4_format },
+  { "ifdef",		false,	false,	true,	m4_ifdef },
+  { "ifelse",		false,	false,	true,	m4_ifelse },
+  { "include",		false,	false,	true,	m4_include },
+  { "incr",		false,	false,	true,	m4_incr },
+  { "index",		false,	false,	true,	m4_index },
+  { "indir",		true,	true,	true,	m4_indir },
+  { "len",		false,	false,	true,	m4_len },
+  { "m4exit",		false,	false,	false,	m4_m4exit },
+  { "m4wrap",		false,	false,	true,	m4_m4wrap },
+  { "maketemp",		false,	false,	true,	m4_maketemp },
+  { "mkstemp",		false,	false,	true,	m4_mkstemp },
+  { "patsubst",		true,	false,	true,	m4_patsubst },
+  { "popdef",		false,	false,	true,	m4_popdef },
+  { "pushdef",		false,	true,	true,	m4_pushdef },
+  { "regexp",		true,	false,	true,	m4_regexp },
+  { "shift",		false,	false,	true,	m4_shift },
+  { "sinclude",		false,	false,	true,	m4_sinclude },
+  { "substr",		false,	false,	true,	m4_substr },
+  { "syscmd",		false,	false,	true,	m4_syscmd },
+  { "sysval",		false,	false,	false,	m4_sysval },
+  { "traceoff",		false,	false,	false,	m4_traceoff },
+  { "traceon",		false,	false,	false,	m4_traceon },
+  { "translit",		false,	false,	true,	m4_translit },
+  { "undefine",		false,	false,	true,	m4_undefine },
+  { "undivert",		false,	false,	false,	m4_undivert },
 
-  { 0,			FALSE,	FALSE,	FALSE,	0 },
+  { 0,			false,	false,	false,	0 },
 
   /* placeholder is intentionally stuck after the table end delimiter,
      so that we can easily find it, while not treating it as a real
      builtin.  */
-  { "placeholder",	TRUE,	FALSE,	FALSE,	m4_placeholder },
+  { "placeholder",	true,	false,	false,	m4_placeholder },
 };
 
 static predefined const
@@ -219,6 +221,68 @@ define_builtin (const char *name, const builtin *bp, symbol_lookup mode)
   SYMBOL_FUNC (sym) = bp->func;
 }
 
+/* Storage for the compiled regular expression of
+   --warn-macro-sequence.  */
+static struct re_pattern_buffer macro_sequence_buf;
+
+/* Storage for the matches of --warn-macro-sequence.  */
+static struct re_registers macro_sequence_regs;
+
+/* True if --warn-macro-sequence is in effect.  */
+static bool macro_sequence_inuse;
+
+/*----------------------------------------.
+| Clean up regular expression variables.  |
+`----------------------------------------*/
+
+static void
+free_pattern_buffer (struct re_pattern_buffer *buf, struct re_registers *regs)
+{
+  regfree (buf);
+  free (regs->start);
+  free (regs->end);
+}
+
+/*-----------------------------------------------------------------.
+| Set the regular expression of --warn-macro-sequence that will be |
+| checked during define and pushdef.  Exit on failure.             |
+`-----------------------------------------------------------------*/
+void
+set_macro_sequence (const char *regexp)
+{
+  const char *msg;
+
+  if (! regexp)
+    regexp = "\\$\\({[^}]*}\\|[0-9][0-9]+\\)";
+  else if (regexp[0] == '\0')
+    {
+      macro_sequence_inuse = false;
+      return;
+    }
+
+  msg = re_compile_pattern (regexp, strlen (regexp), &macro_sequence_buf);
+  if (msg != NULL)
+    {
+      M4ERROR ((EXIT_FAILURE, 0,
+		"--warn-macro-sequence: bad regular expression `%s': %s",
+		regexp, msg));
+    }
+  re_set_registers (&macro_sequence_buf, &macro_sequence_regs,
+		    macro_sequence_regs.num_regs,
+		    macro_sequence_regs.start, macro_sequence_regs.end);
+  macro_sequence_inuse = true;
+}
+
+/*------------------------------------------------------------.
+| Free dynamic memory utilized by the define sequence regular |
+| expression.						      |
+`------------------------------------------------------------*/
+void
+free_macro_sequence (void)
+{
+  free_pattern_buffer (&macro_sequence_buf, &macro_sequence_regs);
+}
+
 /*-------------------------------------------------------------------------.
 | Define a predefined or user-defined macro, with name NAME, and expansion |
 | TEXT.  MODE destinguishes between the "define" and the "pushdef" case.   |
@@ -229,17 +293,48 @@ void
 define_user_macro (const char *name, const char *text, symbol_lookup mode)
 {
   symbol *s;
+  char *defn = xstrdup (text ? text : "");
 
   s = lookup_symbol (name, mode);
   if (SYMBOL_TYPE (s) == TOKEN_TEXT)
     free (SYMBOL_TEXT (s));
 
   SYMBOL_TYPE (s) = TOKEN_TEXT;
-  SYMBOL_TEXT (s) = xstrdup (text);
+  SYMBOL_TEXT (s) = defn;
+
+  /* Implement --warn-macro-sequence.  */
+  if (macro_sequence_inuse && text)
+    {
+      regoff_t offset = 0;
+      size_t len = strlen (defn);
+
+      while ((offset = re_search (&macro_sequence_buf, defn, len, offset,
+				  len - offset, &macro_sequence_regs)) >= 0)
+	{
+	  /* Skip empty matches.  */
+	  if (macro_sequence_regs.start[0] == macro_sequence_regs.end[0])
+	    offset++;
+	  else
+	    {
+	      char tmp;
+	      offset = macro_sequence_regs.end[0];
+	      tmp = defn[offset];
+	      defn[offset] = '\0';
+	      M4ERROR ((warning_status, 0,
+			"Warning: definition of `%s' contains sequence `%s'",
+			name, defn + macro_sequence_regs.start[0]));
+	      defn[offset] = tmp;
+	    }
+	}
+      if (offset == -2)
+	M4ERROR ((warning_status, 0,
+		  "error checking --warn-define-sequence for macro `%s'",
+		  name));
+    }
 }
 
 /*-----------------------------------------------.
-| Initialise all builtin and predefined macros.	 |
+| Initialize all builtin and predefined macros.	 |
 `-----------------------------------------------*/
 
 void
@@ -285,10 +380,10 @@ builtin_init (void)
 | applicable.								  |
 `------------------------------------------------------------------------*/
 
-static boolean
+static bool
 bad_argc (token_data *name, int argc, int min, int max)
 {
-  boolean isbad = FALSE;
+  bool isbad = false;
 
   if (min > 0 && argc < min)
     {
@@ -296,7 +391,7 @@ bad_argc (token_data *name, int argc, int min, int max)
 	M4ERROR ((warning_status, 0,
 		  "Warning: too few arguments to builtin `%s'",
 		  TOKEN_DATA_TEXT (name)));
-      isbad = TRUE;
+      isbad = true;
     }
   else if (max > 0 && argc > max && !suppress_warnings)
     M4ERROR ((warning_status, 0,
@@ -309,10 +404,10 @@ bad_argc (token_data *name, int argc, int min, int max)
 /*--------------------------------------------------------------------------.
 | The function numeric_arg () converts ARG to an int pointed to by VALUEP.  |
 | If the conversion fails, print error message for macro MACRO.  Return	    |
-| TRUE iff conversion succeeds.						    |
+| true iff conversion succeeds.						    |
 `--------------------------------------------------------------------------*/
 
-static boolean
+static bool
 numeric_arg (token_data *macro, const char *arg, int *valuep)
 {
   char *endp;
@@ -333,7 +428,7 @@ numeric_arg (token_data *macro, const char *arg, int *valuep)
 	  M4ERROR ((warning_status, 0,
 		    "non-numeric argument to builtin `%s'",
 		    TOKEN_DATA_TEXT (macro)));
-	  return FALSE;
+	  return false;
 	}
       if (isspace (to_uchar (*arg)))
 	M4ERROR ((warning_status, 0,
@@ -344,7 +439,7 @@ numeric_arg (token_data *macro, const char *arg, int *valuep)
 		  "numeric overflow detected in builtin `%s'",
 		  TOKEN_DATA_TEXT (macro)));
     }
-  return TRUE;
+  return true;
 }
 
 /*------------------------------------------------------------------------.
@@ -356,24 +451,24 @@ numeric_arg (token_data *macro, const char *arg, int *valuep)
 static char const digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
 
 static const char *
-ntoa (register eval_t value, int radix)
+ntoa (int32_t value, int radix)
 {
-  boolean negative;
-  unsigned_eval_t uvalue;
+  bool negative;
+  uint32_t uvalue;
   static char str[256];
-  register char *s = &str[sizeof str];
+  char *s = &str[sizeof str];
 
   *--s = '\0';
 
   if (value < 0)
     {
-      negative = TRUE;
-      uvalue = (unsigned_eval_t) -value;
+      negative = true;
+      uvalue = -(uint32_t) value;
     }
   else
     {
-      negative = FALSE;
-      uvalue = (unsigned_eval_t) value;
+      negative = false;
+      uvalue = (uint32_t) value;
     }
 
   do
@@ -398,18 +493,18 @@ shipout_int (struct obstack *obs, int val)
 {
   const char *s;
 
-  s = ntoa ((eval_t) val, 10);
+  s = ntoa ((int32_t) val, 10);
   obstack_grow (obs, s, strlen (s));
 }
 
 /*----------------------------------------------------------------------.
 | Print ARGC arguments from the table ARGV to obstack OBS, separated by |
-| SEP, and quoted by the current quotes, if QUOTED is TRUE.	        |
+| SEP, and quoted by the current quotes, if QUOTED is true.	        |
 `----------------------------------------------------------------------*/
 
 static void
 dump_args (struct obstack *obs, int argc, token_data **argv,
-	   const char *sep, boolean quoted)
+	   const char *sep, bool quoted)
 {
   int i;
   size_t len = strlen (sep);
@@ -456,7 +551,11 @@ define_macro (int argc, token_data **argv, symbol_lookup mode)
     return;
 
   if (TOKEN_DATA_TYPE (argv[1]) != TOKEN_TEXT)
-    return;
+    {
+      M4ERROR ((warning_status, 0,
+		"Warning: %s: invalid macro name ignored", ARG (0)));
+      return;
+    }
 
   if (argc == 2)
     {
@@ -602,8 +701,9 @@ struct dump_symbol_data
 };
 
 static void
-dump_symbol (symbol *sym, struct dump_symbol_data *data)
+dump_symbol (symbol *sym, void *arg)
 {
+  struct dump_symbol_data *data = (struct dump_symbol_data *) arg;
   if (!SYMBOL_SHADOWED (sym) && SYMBOL_TYPE (sym) != TOKEN_VOID)
     {
       obstack_blank (data->obs, sizeof (symbol *));
@@ -642,7 +742,7 @@ m4_dumpdef (struct obstack *obs, int argc, token_data **argv)
 
   if (argc == 1)
     {
-      hack_all_symbols (dump_symbol, (char *) &data);
+      hack_all_symbols (dump_symbol, &data);
     }
   else
     {
@@ -659,9 +759,9 @@ m4_dumpdef (struct obstack *obs, int argc, token_data **argv)
 
   /* Make table of symbols invisible to expand_macro ().  */
 
-  (void) obstack_finish (obs);
+  obstack_finish (obs);
 
-  qsort ((char *) data.base, data.size, sizeof (symbol *), dumpdef_cmp);
+  qsort (data.base, data.size, sizeof (symbol *), dumpdef_cmp);
 
   for (; data.size > 0; --data.size, data.base++)
     {
@@ -708,17 +808,34 @@ static void
 m4_builtin (struct obstack *obs, int argc, token_data **argv)
 {
   const builtin *bp;
-  const char *name = ARG (1);
+  const char *name;
 
   if (bad_argc (argv[0], argc, 2, -1))
     return;
+  if (TOKEN_DATA_TYPE (argv[1]) != TOKEN_TEXT)
+    {
+      M4ERROR ((warning_status, 0,
+		"Warning: %s: invalid macro name ignored", ARG (0)));
+      return;
+    }
 
+  name = ARG (1);
   bp = find_builtin_by_name (name);
   if (bp->func == m4_placeholder)
     M4ERROR ((warning_status, 0,
 	      "undefined builtin `%s'", name));
   else
-    (*bp->func) (obs, argc - 1, argv + 1);
+    {
+      int i;
+      if (! bp->groks_macro_args)
+	for (i = 2; i < argc; i++)
+	  if (TOKEN_DATA_TYPE (argv[i]) != TOKEN_TEXT)
+	    {
+	      TOKEN_DATA_TYPE (argv[i]) = TOKEN_TEXT;
+	      TOKEN_DATA_TEXT (argv[i]) = (char *) "";
+	    }
+      bp->func (obs, argc - 1, argv + 1);
+    }
 }
 
 /*------------------------------------------------------------------------.
@@ -732,23 +849,40 @@ static void
 m4_indir (struct obstack *obs, int argc, token_data **argv)
 {
   symbol *s;
-  const char *name = ARG (1);
+  const char *name;
 
   if (bad_argc (argv[0], argc, 2, -1))
     return;
+  if (TOKEN_DATA_TYPE (argv[1]) != TOKEN_TEXT)
+    {
+      M4ERROR ((warning_status, 0,
+		"Warning: %s: invalid macro name ignored", ARG (0)));
+      return;
+    }
 
+  name = ARG (1);
   s = lookup_symbol (name, SYMBOL_LOOKUP);
   if (s == NULL || SYMBOL_TYPE (s) == TOKEN_VOID)
     M4ERROR ((warning_status, 0,
 	      "undefined macro `%s'", name));
   else
-    call_macro (s, argc - 1, argv + 1, obs);
+    {
+      int i;
+      if (! SYMBOL_MACRO_ARGS (s))
+	for (i = 2; i < argc; i++)
+	  if (TOKEN_DATA_TYPE (argv[i]) != TOKEN_TEXT)
+	    {
+	      TOKEN_DATA_TYPE (argv[i]) = TOKEN_TEXT;
+	      TOKEN_DATA_TEXT (argv[i]) = (char *) "";
+	    }
+      call_macro (s, argc - 1, argv + 1, obs);
+    }
 }
 
 /*-------------------------------------------------------------------------.
 | The macro "defn" returns the quoted definition of the macro named by the |
 | first argument.  If the macro is builtin, it will push a special	   |
-| macro-definition token on ht input stack.				   |
+| macro-definition token on the input stack.				   |
 `-------------------------------------------------------------------------*/
 
 static void
@@ -898,7 +1032,7 @@ m4_sysval (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_eval (struct obstack *obs, int argc, token_data **argv)
 {
-  eval_t value = 0;
+  int32_t value = 0;
   int radix = 10;
   int min = 1;
   const char *s;
@@ -912,8 +1046,8 @@ m4_eval (struct obstack *obs, int argc, token_data **argv)
   if (radix < 1 || radix > (int) strlen (digits))
     {
       M4ERROR ((warning_status, 0,
-		"radix in builtin `%s' out of range (radix = %d)",
-		ARG (0), radix));
+		"radix %d in builtin `%s' out of range",
+		radix, ARG (0)));
       return;
     }
 
@@ -1050,11 +1184,13 @@ m4_undivert (struct obstack *obs, int argc, token_data **argv)
 		    "non-numeric argument to builtin `%s'", ARG (0)));
 	else
 	  {
-	    fp = path_search (ARG (i), NULL);
+	    fp = m4_path_search (ARG (i), NULL);
 	    if (fp != NULL)
 	      {
 		insert_file (fp);
-		fclose (fp);
+		if (fclose (fp) == EOF)
+		  M4ERROR ((warning_status, errno,
+			    "error undiverting `%s'", ARG (i)));
 	      }
 	    else
 	      M4ERROR ((warning_status, errno,
@@ -1091,7 +1227,7 @@ m4_shift (struct obstack *obs, int argc, token_data **argv)
 {
   if (bad_argc (argv[0], argc, 2, -1))
     return;
-  dump_args (obs, argc - 1, argv + 1, ",", TRUE);
+  dump_args (obs, argc - 1, argv + 1, ",", true);
 }
 
 /*--------------------------------------------------------------------------.
@@ -1104,6 +1240,7 @@ m4_changequote (struct obstack *obs, int argc, token_data **argv)
   if (bad_argc (argv[0], argc, 1, 3))
     return;
 
+  /* Explicit NULL distinguishes between empty and missing argument.  */
   set_quotes ((argc >= 2) ? TOKEN_DATA_TEXT (argv[1]) : NULL,
 	     (argc >= 3) ? TOKEN_DATA_TEXT (argv[2]) : NULL);
 }
@@ -1119,11 +1256,9 @@ m4_changecom (struct obstack *obs, int argc, token_data **argv)
   if (bad_argc (argv[0], argc, 1, 3))
     return;
 
-  if (argc == 1)
-    set_comment ("", "");	/* disable comments */
-  else
-    set_comment (TOKEN_DATA_TEXT (argv[1]),
-		(argc >= 3) ? TOKEN_DATA_TEXT (argv[2]) : NULL);
+  /* Explicit NULL distinguishes between empty and missing argument.  */
+  set_comment ((argc >= 2) ? TOKEN_DATA_TEXT (argv[1]) : NULL,
+	       (argc >= 3) ? TOKEN_DATA_TEXT (argv[2]) : NULL);
 }
 
 #ifdef ENABLE_CHANGEWORD
@@ -1150,29 +1285,31 @@ m4_changeword (struct obstack *obs, int argc, token_data **argv)
 
 /*-------------------------------------------------------------------------.
 | Generic include function.  Include the file given by the first argument, |
-| if it exists.  Complain about inaccesible files iff SILENT is FALSE.	   |
+| if it exists.  Complain about inaccesible files iff SILENT is false.	   |
 `-------------------------------------------------------------------------*/
 
 static void
-include (int argc, token_data **argv, boolean silent)
+include (int argc, token_data **argv, bool silent)
 {
   FILE *fp;
-  const char *name;
+  char *name;
 
   if (bad_argc (argv[0], argc, 2, 2))
     return;
 
-  fp = path_search (ARG (1), &name);
+  fp = m4_path_search (ARG (1), &name);
   if (fp == NULL)
     {
       if (!silent)
-	M4ERROR ((warning_status, errno,
-		  "cannot open `%s'", ARG (1)));
+	{
+	  M4ERROR ((warning_status, errno, "cannot open `%s'", ARG (1)));
+	  retcode = EXIT_FAILURE;
+	}
       return;
     }
 
-  push_file (fp, name);
-  free ((char *) name);
+  push_file (fp, name, true);
+  free (name);
 }
 
 /*------------------------------------------------.
@@ -1182,7 +1319,7 @@ include (int argc, token_data **argv, boolean silent)
 static void
 m4_include (struct obstack *obs, int argc, token_data **argv)
 {
-  include (argc, argv, FALSE);
+  include (argc, argv, false);
 }
 
 /*----------------------------------.
@@ -1192,7 +1329,7 @@ m4_include (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_sinclude (struct obstack *obs, int argc, token_data **argv)
 {
-  include (argc, argv, TRUE);
+  include (argc, argv, true);
 }
 
 /* More miscellaneous builtins -- "maketemp", "errprint", "__file__",
@@ -1202,21 +1339,86 @@ m4_sinclude (struct obstack *obs, int argc, token_data **argv)
 | Use the first argument as at template for a temporary file name.  |
 `------------------------------------------------------------------*/
 
+/* Add trailing 'X' to NAME if necessary, securely create the file,
+   and place the new file name on OBS.  */
+static void
+mkstemp_helper (struct obstack *obs, const char *name)
+{
+  int fd;
+  int len;
+  int i;
+
+  /* Guarantee that there are six trailing 'X' characters, even if the
+     user forgot to supply them.  */
+  len = strlen (name);
+  obstack_grow (obs, name, len);
+  for (i = 0; len > 0 && i < 6; i++)
+    if (name[--len] != 'X')
+      break;
+  for (; i < 6; i++)
+    obstack_1grow (obs, 'X');
+  obstack_1grow (obs, '\0');
+
+  errno = 0;
+  fd = mkstemp ((char *) obstack_base (obs));
+  if (fd < 0)
+    {
+      M4ERROR ((0, errno, "cannot create tempfile `%s'", name));
+      obstack_free (obs, obstack_finish (obs));
+    }
+  else
+    close (fd);
+}
+
 static void
 m4_maketemp (struct obstack *obs, int argc, token_data **argv)
 {
-  int fd;
   if (bad_argc (argv[0], argc, 2, 2))
     return;
-  errno = 0;
-  if ((fd = mkstemp (ARG (1))) < 0)
+  if (no_gnu_extensions)
     {
-      M4ERROR ((warning_status, errno, "cannot create tempfile `%s'",
-		ARG (1)));
-      return;
+      /* POSIX states "any trailing 'X' characters [are] replaced with
+	 the current process ID as a string", without referencing the
+	 file system.  Horribly insecure, but we have to do it when we
+	 are in traditional mode.
+
+	 For reference, Solaris m4 does:
+	   maketemp() -> `'
+	   maketemp(X) -> `X'
+	   maketemp(XX) -> `Xn', where n is last digit of pid
+	   maketemp(XXXXXXXX) -> `X00nnnnn', where nnnnn is 16-bit pid
+      */
+      const char *str = ARG (1);
+      int len = strlen (str);
+      int i;
+      int len2;
+
+      M4ERROR ((warning_status, 0, "recommend using mkstemp instead"));
+      for (i = len; i > 1; i--)
+	if (str[i - 1] != 'X')
+	  break;
+      obstack_grow (obs, str, i);
+      str = ntoa ((int32_t) getpid (), 10);
+      len2 = strlen (str);
+      if (len2 > len - i)
+	obstack_grow0 (obs, str + len2 - (len - i), len - i);
+      else
+	{
+	  while (i++ < len - len2)
+	    obstack_1grow (obs, '0');
+	  obstack_grow0 (obs, str, len2);
+	}
     }
-  close(fd);
-  obstack_grow (obs, ARG (1), strlen (ARG (1)));
+  else
+    mkstemp_helper (obs, ARG (1));
+}
+
+static void
+m4_mkstemp (struct obstack *obs, int argc, token_data **argv)
+{
+  if (bad_argc (argv[0], argc, 2, 2))
+    return;
+  mkstemp_helper (obs, ARG (1));
 }
 
 /*----------------------------------------.
@@ -1228,7 +1430,7 @@ m4_errprint (struct obstack *obs, int argc, token_data **argv)
 {
   if (bad_argc (argv[0], argc, 2, -1))
     return;
-  dump_args (obs, argc, argv, " ", FALSE);
+  dump_args (obs, argc, argv, " ", false);
   obstack_1grow (obs, '\0');
   debug_flush_files ();
   fprintf (stderr, "%s", (char *) obstack_finish (obs));
@@ -1287,17 +1489,15 @@ m4_m4exit (struct obstack *obs, int argc, token_data **argv)
 		"exit status out of range: `%d'", exit_code));
       exit_code = EXIT_FAILURE;
     }
-  if (close_stream (stdout) != 0)
-    {
-      M4ERROR ((warning_status, errno, "write error"));
-      if (exit_code == 0)
-	exit_code = EXIT_FAILURE;
-    }
   /* Change debug stream back to stderr, to force flushing debug stream and
      detect any errors it might have encountered.  */
   debug_set_output (NULL);
-  if (exit_code == 0 && retcode != 0)
+  debug_flush_files ();
+  if (exit_code == EXIT_SUCCESS && retcode != EXIT_SUCCESS)
     exit_code = retcode;
+  /* Propagate non-zero status to atexit handlers.  */
+  if (exit_code != EXIT_SUCCESS)
+    exit_failure = exit_code;
   exit (exit_code);
 }
 
@@ -1315,9 +1515,9 @@ m4_m4wrap (struct obstack *obs, int argc, token_data **argv)
   if (no_gnu_extensions)
     obstack_grow (obs, ARG (1), strlen (ARG (1)));
   else
-    dump_args (obs, argc, argv, " ", FALSE);
+    dump_args (obs, argc, argv, " ", false);
   obstack_1grow (obs, '\0');
-  push_wrapup (obstack_finish (obs));
+  push_wrapup ((char *) obstack_finish (obs));
 }
 
 /* Enable tracing of all specified macros, or all, if none is specified.
@@ -1331,9 +1531,9 @@ m4_m4wrap (struct obstack *obs, int argc, token_data **argv)
 `-----------------------------------------------------------------------*/
 
 static void
-set_trace (symbol *sym, const char *data)
+set_trace (symbol *sym, void *data)
 {
-  SYMBOL_TRACED (sym) = (boolean) (data != NULL);
+  SYMBOL_TRACED (sym) = data != NULL;
   /* Remove placeholder from table if macro is undefined and untraced.  */
   if (SYMBOL_TYPE (sym) == TOKEN_VOID && data == NULL)
     lookup_symbol (SYMBOL_NAME (sym), SYMBOL_POPDEF);
@@ -1346,12 +1546,12 @@ m4_traceon (struct obstack *obs, int argc, token_data **argv)
   int i;
 
   if (argc == 1)
-    hack_all_symbols (set_trace, (char *) obs);
+    hack_all_symbols (set_trace, obs);
   else
     for (i = 1; i < argc; i++)
       {
 	s = lookup_symbol (TOKEN_DATA_TEXT (argv[i]), SYMBOL_INSERT);
-	set_trace (s, (char *) obs);
+	set_trace (s, obs);
       }
 }
 
@@ -1471,8 +1671,9 @@ m4_len (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_index (struct obstack *obs, int argc, token_data **argv)
 {
-  const char *cp, *last;
-  int l1, l2, retval;
+  const char *haystack;
+  const char *result;
+  int retval;
 
   if (bad_argc (argv[0], argc, 3, 3))
     {
@@ -1482,17 +1683,9 @@ m4_index (struct obstack *obs, int argc, token_data **argv)
       return;
     }
 
-  l1 = strlen (ARG (1));
-  l2 = strlen (ARG (2));
-
-  last = ARG (1) + l1 - l2;
-
-  for (cp = ARG (1); cp <= last; cp++)
-    {
-      if (strncmp (cp, ARG (2), l2) == 0)
-	break;
-    }
-  retval = (cp <= last) ? cp - ARG (1) : -1;
+  haystack = ARG (1);
+  result = strstr (haystack, ARG (2));
+  retval = result ? result - haystack : -1;
 
   shipout_int (obs, retval);
 }
@@ -1546,14 +1739,14 @@ m4_substr (struct obstack *obs, int argc, token_data **argv)
 static const char *
 expand_ranges (const char *s, struct obstack *obs)
 {
-  char from;
-  char to;
+  unsigned char from;
+  unsigned char to;
 
-  for (from = '\0'; *s != '\0'; from = *s++)
+  for (from = '\0'; *s != '\0'; from = to_uchar (*s++))
     {
       if (*s == '-' && from != '\0')
 	{
-	  to = *++s;
+	  to = to_uchar (*++s);
 	  if (to == '\0')
 	    {
 	      /* trailing dash */
@@ -1575,7 +1768,7 @@ expand_ranges (const char *s, struct obstack *obs)
 	obstack_1grow (obs, *s);
     }
   obstack_1grow (obs, '\0');
-  return obstack_finish (obs);
+  return (char *) obstack_finish (obs);
 }
 
 /*----------------------------------------------------------------------.
@@ -1589,9 +1782,12 @@ expand_ranges (const char *s, struct obstack *obs)
 static void
 m4_translit (struct obstack *obs, int argc, token_data **argv)
 {
-  register const char *data, *tmp;
-  const char *from, *to;
-  int tolen;
+  const char *data;
+  const char *from;
+  const char *to;
+  char map[256] = {0};
+  char found[256] = {0};
+  unsigned char ch;
 
   if (bad_argc (argv[0], argc, 3, 4))
     {
@@ -1609,33 +1805,37 @@ m4_translit (struct obstack *obs, int argc, token_data **argv)
 	return;
     }
 
-  if (argc >= 4)
+  to = ARG (3);
+  if (strchr (to, '-') != NULL)
     {
-      to = ARG (3);
-      if (strchr (to, '-') != NULL)
-	{
-	  to = expand_ranges (to, obs);
-	  if (to == NULL)
-	    return;
-	}
+      to = expand_ranges (to, obs);
+      if (to == NULL)
+	return;
     }
-  else
-    to = "";
 
-  tolen = strlen (to);
-
-  for (data = ARG (1); *data; data++)
+  /* Calling strchr(from) for each character in data is quadratic,
+     since both strings can be arbitrarily long.  Instead, create a
+     from-to mapping in one pass of from, then use that map in one
+     pass of data, for linear behavior.  Traditional behavior is that
+     only the first instance of a character in from is consulted,
+     hence the found map.  */
+  for ( ; (ch = *from) != '\0'; from++)
     {
-      tmp = strchr (from, *data);
-      if (tmp == NULL)
+      if (! found[ch])
 	{
-	  obstack_1grow (obs, *data);
-	}
-      else
-	{
-	  if (tmp - from < tolen)
-	    obstack_1grow (obs, *(to + (tmp - from)));
+	  found[ch] = 1;
+	  map[ch] = *to;
 	}
+      if (*to != '\0')
+	to++;
+    }
+
+  for (data = ARG (1); (ch = *data) != '\0'; data++)
+    {
+      if (! found[ch])
+	obstack_1grow (obs, ch);
+      else if (map[ch])
+	obstack_1grow (obs, map[ch]);
     }
 }
 
@@ -1667,7 +1867,7 @@ static void
 substitute (struct obstack *obs, const char *victim, const char *repl,
 	    struct re_registers *regs)
 {
-  register unsigned int ch;
+  int ch;
 
   for (;;)
     {
@@ -1721,27 +1921,18 @@ Warning: trailing \\ ignored in replacement"));
 | Initialize regular expression variables.  |
 `------------------------------------------*/
 
-static void
+void
 init_pattern_buffer (struct re_pattern_buffer *buf, struct re_registers *regs)
 {
   buf->translate = NULL;
   buf->fastmap = NULL;
   buf->buffer = NULL;
   buf->allocated = 0;
-  regs->start = NULL;
-  regs->end = NULL;
-}
-
-/*----------------------------------------.
-| Clean up regular expression variables.  |
-`----------------------------------------*/
-
-static void
-free_pattern_buffer (struct re_pattern_buffer *buf, struct re_registers *regs)
-{
-  regfree (buf);
-  free (regs->start);
-  free (regs->end);
+  if (regs)
+    {
+      regs->start = NULL;
+      regs->end = NULL;
+    }
 }
 
 /*--------------------------------------------------------------------------.
@@ -1926,7 +2117,7 @@ void
 expand_user_macro (struct obstack *obs, symbol *sym,
 		   int argc, token_data **argv)
 {
-  register const char *text;
+  const char *text;
   int i;
 
   for (text = SYMBOL_TEXT (sym); *text != '\0';)
